from translatepy import Translator
from translatepy.exceptions import TranslatepyException, UnknownLanguage

from .base import Translator as BaseTranslator


class TranslatePy(BaseTranslator):
    max_char = 1e10

    def __init__(self):
        self.translator = Translator()

    def translate(self, text, source_language, destination_language):
        result = self.translator.translate(
            text,
            source_language=source_language,
            destination_language=destination_language,
        )
        return result.result



import deepl
from .base import Translator


class DeeplApi(Translator):
    max_char = 1500

    def __init__(self, api_key):
        self.translator = deepl.Translator(api_key)

    def translate(self, text: str, source_language: str, destination_language: str):
        result = self.translator.translate_text(
            text, source_lang=source_language, target_lang=destination_language
        )
        return result.text



from abc import ABC, abstractmethod


class Translator(ABC):
    max_char: int

    @abstractmethod
    def translate(
        self, text: str, source_language: str, destination_language: str
    ) -> str:
        ...

    def quit(self):
        ...


class TimeOutException(Exception):
    """Translation timed out"""



import time
import logging

from typing import Optional
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.common.proxy import Proxy
from selenium.webdriver.common.keys import Keys

from .base import Translator, TimeOutException
from .selenium_utils import (
    create_proxy,
    create_driver,
    TextArea,
    Button,
    Text,
)


class DeeplTranslator(Translator):
    url = "https://www.deepl.com/translator"
    max_char = 1500
    languages = {
        "auto": "Any language (detect)",
        "bg": "Bulgarian",
        "zh": "Chinese",
        "cs": "Czech",
        "da": "Danish",
        "nl": "Dutch",
        "en": "English",  # Only usable for source language
        "en-US": "English (American)",  # Only usable for destination language
        "en-GB": "English (British)",  # Only usable for destination language
        "et": "Estonian",
        "fi": "Finnish",
        "fr": "French",
        "de": "German",
        "el": "Greek",
        "hu": "Hungarian",
        "id": "Indonesian",
        "it": "Italian",
        "ja": "Japanese",
        "ko": "Korean",
        "lv": "Latvian",
        "lt": "Lithuanian",
        "pl": "Polish",
        "pt": "Portuguese",  # Only usable for source language
        "pt-PT": "Portuguese",  # Only usable for destination language
        "pt-BR": "Portuguese (Brazilian)",  # Only usable for destination language
        "ro": "Romanian",
        "ru": "Russian",
        "sk": "Slovak",
        "sl": "Slovenian",
        "es": "Spanish",
        "sv": "Swedish",
        "tr": "Turkish",
        "uk": "Ukrainian",
    }

    def __init__(self, driver: Optional[WebDriver] = None):
        self.last_translation_failed = False
        self.driver = driver

        if self.driver is None:
            self._rotate_proxy()
            return

        self._reset()

    def _reset(self):
        logging.info(f"Going to {self.url}")
        self.driver.get(self.url)
        self.driver.implicitly_wait(5)
        self._closePopUp()

        self.input_lang_from = TextArea(
            self.driver, "XPATH", "//*[@data-testid='translator-source-input']"
        )
        self.input_destination_language = TextArea(
            self.driver, "XPATH", "//*[@data-testid='translator-target-input']"
        )

        self.src_lang = None
        self.target_lang = None

    def _rotate_proxy(self):
        if self.driver is not None:
            logging.info(" ======= Translation failed. Probably got banned. ======= ")
            logging.info("Rotating proxy")
            self.quit()

        proxy = create_proxy()
        self.driver = create_driver(proxy)
        self._reset()

    def _closePopUp(self):
        Button(
            self.driver,
            "CSS_SELECTOR",
            "[aria-label=Close]",
            wait_time=5,
            optional=True,
        ).click()

    def _set_source_language(self, language: str) -> None:
        self._set_language(language, "//*[@data-testid='translator-source-lang']")
        self.src_lang = language

    def _set_destination_language(self, language: str) -> None:
        self._set_language(language, "//*[@data-testid='translator-target-lang']")
        self.target_lang = language

    def _set_language(self, language: str, xpath: str) -> None:
        # Click the languages dropdown button
        Button(self.driver, "XPATH", xpath).click()

        # Get the language button to click based on is dl-test property or the text in the button
        xpath_by_property = (
            f"//button[@data-testid='translator-lang-option-{language}']"
        )
        x_path_by_text = f"//button[text()='{self.languages[language]}']"
        xpath = f"{xpath_by_property} | {x_path_by_text}"

        # Click the wanted language button
        Button(self.driver, "XPATH", xpath).click()

    def _is_translated(self, original: str, translation: str) -> bool:
        return (
            len(translation) != 0
            and "[...]" not in translation
            and len(original.splitlines()) == len(translation.splitlines())
            and original != translation
        )

    def translate(self, text: str, source_language: str, destination_language: str):
        if source_language != self.src_lang:
            self._set_source_language(source_language)
        if destination_language != self.target_lang:
            self._set_destination_language(destination_language)

        clean_text = text.replace("[...]", "@[.]@")

        self.input_lang_from.write((clean_text.replace("\n", Keys.ENTER)))

        # Maximun number of iterations 60 seconds
        for _ in range(60):
            translation = self.input_destination_language.value

            if self._is_translated(clean_text, translation):
                time.sleep(2)
                translation = self.input_destination_language.value

                # Reset the proxy flag
                self.last_translation_failed = False
                return translation.replace("@[.]@", "[...]")
            time.sleep(1)

        # Maybe proxy got banned, so we try with a new proxy, but just once.
        if not self.last_translation_failed:
            self.last_translation_failed = True
            self._rotate_proxy()
            return self.translate(text, source_language, destination_language)

        self.quit()
        raise TimeOutException("Translation timed out")

    def quit(self):
        self.driver.quit()



import os
import re
import pyass

from typing import List, Generator

from .translators.base import Translator


class AssFile:
    """ASS file class abstraction

    Args:
        filepath (str): file path of ass
    """

    def __init__(self, filepath: str) -> None:
        self.filepath = filepath
        self.backup_file = f"{self.filepath}.tmp"
        self.subtitles = []
        self.start_from = 0
        self.current_subtitle = 0
        self.text_styles = []

        print(f"Loading {filepath} as ASS")
        with open(filepath, "r", encoding="utf-8", errors="ignore") as input_file:
            self.subtitles = self.load_from_file(input_file)

        self._load_backup()

    def _load_backup(self):
        if not os.path.exists(self.backup_file):
            return

        print(f"Backup file found = {self.backup_file}")
        with open(
            self.backup_file, "r", encoding="utf-8", errors="ignore"
        ) as input_file:
            subtitles = self.load_from_file(input_file)

            self.start_from = len(subtitles.events)
            self.current_subtitle = self.start_from
            print(f"Starting from subtitle {self.start_from}")
            self.subtitles.events = [
                *subtitles.events,
                *self.subtitles.events[self.start_from :],
            ]

    def load_from_file(self, input_file):
        ass_file        = pyass.load(input_file)
        ass_file.events = sorted(ass_file.events, key=lambda e: (e.start))
        return self._clean_subs_content(ass_file)

    def _get_next_chunk(self, chunk_size: int = 4500) -> Generator:
        """Get a portion of the subtitles at the time based on the chunk size

        Args:
            chunk_size (int, optional): Maximum number of letter in text chunk. Defaults to 4500.

        Yields:
            Generator: Each chunk at the time
        """
        portion = []

        for subtitle in self.subtitles.events[self.start_from :]:
            # Manage ASS styles for subtitle before add it to the portion
            # Extract a list of styles
            # Replace the styles by |

            # Each style starts with { and end with }
            # If we have an "}" then we can split and keep the part on the left and keep it in our list
            for i in subtitle.text.split("{"):
                if "}" in i:
                    self.text_styles.append("{" + i.split("}")[0] + "}")

            subtitle.text = re.sub(r"{.*?}", r"|", subtitle.text)

            # Calculate new chunk size if subtitle content is added to actual chunk
            n_char = (
                sum(len(sub.text) for sub in portion)  # All subtitles in chunk
                + len(subtitle.text)  # New subtitle
                + len(portion)  # Break lines in chunk
                + 1  # New breakline
            )

            # If chunk goes beyond the limit, yield it
            if n_char >= chunk_size and len(portion) != 0:
                yield portion
                portion = []

            # Put subtitle content in chunk
            portion.append(subtitle)

        # Yield last chunk
        yield portion

    def _clean_subs_content(self, subtitles):
        """Cleans subtitles content and delete line breaks

        Args:
            subtitles List of subtitles

        Returns:
            Same list of subtitles, but cleaned
        """
        cleanr = re.compile("<.*?>")

        for sub in subtitles.events:
            sub.text = cleanr.sub("", sub.text)
            # No real equivalent in ASS
            #sub.text = srt.make_legal_content(sub.content)
            sub.text = sub.text.strip()

            if sub.text == "":
                sub.text = "..."

            if all(sentence.startswith("-") for sentence in sub.text.split("\n")):
                sub.text = sub.text.replace("\n", "////")
                continue

            # It looks like \N is removed by the translation so we replace them by \\\\
            sub.text = sub.text.replace(r"\N", r"\\\\")

            # The \\\\ must be separated from the words to avoid weird conversions
            sub.text = re.sub(r"[aA0-zZ9]\\\\", r" \\\\", sub.text)
            sub.text = re.sub(r"\\\\[aA0-zZ9]", r"\\\\ ", sub.text)

            sub.text = sub.text.replace("\n", " ")

        return subtitles

    def wrap_lines(self, line_wrap_limit: int = 50) -> None:
        """

        Args:
            line_wrap_limit (int): Number of maximum characters in a line before wrap. Defaults to 50. (not used)
        """
        for sub in self.subtitles.events:
            sub.text = sub.text.replace("////", "\n")
            sub.text = sub.text.replace(r" \\\\ ", r"\N")

    def translate(
        self,
        translator: Translator,
        source_language: str,
        destination_language: str,
    ) -> None:
        """Translate ASS file using a translator of your choose

        Args:
            translator (Translator): Translator object of choose
            destination_language (str): Destination language (must be coherent with your translator)
            source_language (str): Source language (must be coherent with your translator)
        """

        # For each chunk of the file (based on the translator capabilities)
        for subs_slice in self._get_next_chunk(translator.max_char):
            progress = int(100 * self.current_subtitle / len(self.subtitles.events))
            print(f"... Translating {progress} %")

            # Put chunk in a single text with break lines
            text = [sub.text for sub in subs_slice]
            text = "\n".join(text)

            # Translate
            translation = translator.translate(
                text, source_language, destination_language
            )

            # Manage ASS commands
            # Insert the styles back in the text instead of |
            self.text_styles.reverse()
            translation_with_styles = ""
            for i in translation.split(r"|"):
                try:
                    # We set i at the left part because the style must "replace" the "|"
                    translation_with_styles += i + self.text_styles.pop()
                except IndexError:
                    translation_with_styles += i

            # Break each line back into subtitle content
            translation = translation_with_styles.splitlines()
            for i in range(len(subs_slice)):
                subs_slice[i].text = translation[i]
                self.current_subtitle += 1

        print(f"... Translation done")

    def save_backup(self):
        self.subtitles.events = self.subtitles.events[: self.current_subtitle]
        self.save(self.backup_file)

    def _delete_backup(self):
        if os.path.exists(self.backup_file):
            os.remove(self.backup_file)

    def save(self, filepath: str) -> None:
        """Saves ASS to file

        Args:
            filepath (str): Path of the new file
        """
        self._delete_backup()

        print(f"Saving {filepath}")
        with open(filepath, "w", encoding="utf-8") as file_out:
            pyass.dump(self.subtitles, file_out)



from PyDeepLX import PyDeepLX as PDLX
from random import randint
from time import sleep

from .base import Translator as BaseTranslator
from fp.fp import FreeProxy


class PyDeepLX(BaseTranslator):
    max_char = 1500

    def __init__(self, proxies=None):
        self.proxies = proxies

        # Use proxy by default if self.proxies is True
        if self.proxies:
            print("...... Use proxy")
            self.proxies = FreeProxy(rand=True, timeout=1).get()

    def translate(self, text, source_language, destination_language):
        # Sleep a random number of seconds (between 5 and 10)
        # https://www.shellhacks.com/python-sleep-random-time-web-scraping/
        RANDOM_WAIT = randint(5, 10)
        print(f"...... Wait randomly {RANDOM_WAIT}s")
        sleep(RANDOM_WAIT)

        # Max retry 10
        RETRY_COUNTER = 10
        result = None

        while RETRY_COUNTER > 0 :
            try:
                result = PDLX.translate(
                    text,
                    source_language,
                    destination_language,
                    proxies=self.proxies
                )

                if result == None:
                  print("...... Exception: result is empty raise exception")
                  raise Exception("Result is empty")

                # Everyting alright
                break
            except Exception as e:
                print(f"...... Exception {e} with retry number {RETRY_COUNTER}")

                # Get a random proxy
                print("...... Use or change proxy")
                self.proxies = FreeProxy(rand=True, timeout=1).get()

                # Decrease RETRY_COUNTER
                RETRY_COUNTER -= 1

                # Raise error if RETRY_COUNTER is 0
                if RETRY_COUNTER == 0:
                    print("...... Exception RETRY_COUNTER reached 0")
                    raise

        return result



import sys
import logging

from typing import Optional, List
from fp.fp import FreeProxy
from selenium import webdriver
from webdriverdownloader import GeckoDriverDownloader
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.by import By
from selenium.webdriver.common.proxy import Proxy, ProxyType
from selenium.common.exceptions import WebDriverException
from selenium.webdriver import ActionChains, Keys
from selenium.webdriver.support import expected_conditions as EC


def create_proxy(country_id: Optional[List[str]] = ["US"]) -> Proxy:
    """Creates a new proxy to use with a selenium driver and avoid get banned

    Args:
        country_id (Optional[List[str]], optional): Contry id to create proxy. Defaults to ['US'].

    Returns:
        Proxy: Selenium WebDriver proxy
    """
    i = 0
    while i < 3:
        try:
            logging.info("Getting a new Proxy from https://www.sslproxies.org/")
            proxy = FreeProxy(country_id=country_id).get()
            proxy = Proxy(
                dict(
                    proxyType=ProxyType.MANUAL,
                    httpProxy=proxy,
                    ftpProxy=proxy,
                    sslProxy=proxy,
                    noProxy="",
                )
            )
            return proxy
        except:
            logging.info("Exception while getting Proxy. Trying again")
            i += 1

    raise Exception("Unable to get proxy")


def create_driver(proxy: Optional[Proxy] = None) -> WebDriver:
    """Creates a new Firefox selenium webdriver. Install geckodriver if not in path

    Args:
        proxy (Optional[Proxy], optional): Selenium WebDriver proxy. Defaults to None.

    Returns:
        WebDriver: Selenium WebDriver
    """
    logging.info("Creating Selenium Webdriver instance")
    try:
        driver = webdriver.Firefox(proxy=proxy)
    except WebDriverException:
        logging.info("Installing Firefox GeckoDriver cause it isn't installed")
        gdd = GeckoDriverDownloader()
        gdd.download_and_install()

        driver = webdriver.Firefox(proxy=proxy)

    driver.maximize_window()
    return driver


class BaseElement:
    def __init__(
        self,
        driver: webdriver,
        locate_by: str,
        locate_value: str,
        multiple: bool = False,
        wait_time: int = 100,
        optional: bool = False,
    ) -> None:

        self.driver = driver
        locator = (getattr(By, locate_by.upper(), "id"), locate_value)
        find_element = driver.find_elements if multiple else driver.find_element
        try:
            WebDriverWait(driver, wait_time).until(
                lambda driver: EC.element_to_be_clickable(locator)
            )
            self.element = find_element(*locator)
        except:
            if optional:
                self.element = None
                return

            print(f"Timed out trying to get element ({locate_by} = {locate_value})")
            logging.info("Closing browser")
            driver.quit()
            sys.exit()


class Text(BaseElement):
    @property
    def text(self) -> str:
        if self.element is None:
            return ""

        return self.element.get_attribute("text")


class TextArea(BaseElement):
    def write(self, value: str) -> None:
        if self.element is None:
            return

        # Check OS to use Cmd or Ctrl keys
        cmd_ctrl = Keys.COMMAND if sys.platform == "darwin" else Keys.CONTROL

        actions_handler = ActionChains(self.driver).move_to_element(self.element)
        actions_handler.click().key_down(cmd_ctrl).send_keys("a").perform()
        actions_handler.send_keys(Keys.CLEAR).key_up(cmd_ctrl).perform()
        actions_handler.send_keys(*value).perform()

    @property
    def value(self) -> None:
        if self.element is None:
            return ""

        return self.element.get_attribute("value")


class Button(BaseElement):
    def click(self) -> None:
        if self.element is None:
            return

        try:
            can_click = getattr(self.element, "click", None)
            if callable(can_click):
                self.element.click()
        except:
            # Using javascript if usual click function does not work
            self.driver.execute_script("arguments[0].click();", self.element)



import os
import argparse
import logging
import traceback

from .srt_file import SrtFile
from .ass_file import AssFile
from .translators.deepl_api import DeeplApi
from .translators.deepl_scrap import DeeplTranslator
from .translators.translatepy import TranslatePy
from .translators.pydeeplx import PyDeepLX

parser = argparse.ArgumentParser(description="Translate an .STR and .ASS file")

parser.add_argument(
    "filepath",
    metavar="path",
    type=str,
    help="File to translate",
)

parser.add_argument(
    "-i",
    "--src-lang",
    type=str,
    default="auto",
    help="Source language. Default: auto",
)

parser.add_argument(
    "-o",
    "--dest-lang",
    type=str,
    default="es",
    help="Destination language. Default: es (spanish)",
)

parser.add_argument(
    "-v",
    "--verbose",
    action="store_const",
    dest="loglevel",
    const=logging.INFO,
    help="Increase output verbosity",
)

parser.add_argument(
    "-vv",
    "--debug",
    action="store_const",
    dest="loglevel",
    const=logging.DEBUG,
    default=logging.WARNING,
    help="Increase output verbosity for debugging",
)

parser.add_argument(
    "-s",
    "--show-browser",
    action="store_true",
    help="Show browser window",
)

parser.add_argument(
    "-w",
    "--wrap-limit",
    type=int,
    default=50,
    help="Number of characters -including spaces- to wrap a line of text. Default: 50",
)

parser.add_argument(
    "-t",
    "--translator",
    type=str,
    choices=["deepl-scrap", "translatepy", "deepl-api", "pydeeplx"],
    help="Built-in translator to use",
    default="deepl-scrap",
)

parser.add_argument(
    "--auth",
    type=str,
    help="Api key if needed on translator",
)

parser.add_argument(
    "--proxies",
    action="store_true",
    help="Use proxy by default for pydeeplx",
)

builtin_translators = {
    "deepl-scrap": DeeplTranslator,
    "deepl-api": DeeplApi,
    "translatepy": TranslatePy,
    "pydeeplx": PyDeepLX,
}

args = parser.parse_args()
logging.basicConfig(level=args.loglevel)

try:
    os.environ.pop("MOZ_HEADLESS")
except:
    pass

if not args.show_browser:
    os.environ["MOZ_HEADLESS"] = "1"

translator_args = {}
if args.auth:
    translator_args["api_key"] = args.auth
if args.proxies:
    translator_args["proxies"] = args.proxies    

translator = builtin_translators[args.translator](**translator_args)

try:
    sub = AssFile(args.filepath)
except AttributeError:
    print("... Exception while loading as ASS try as SRT")
    sub = SrtFile(args.filepath)

try:
    sub.translate(translator, args.src_lang, args.dest_lang)
    sub.wrap_lines(args.wrap_limit)
    sub.save(f"{os.path.splitext(args.filepath)[0]}_{args.dest_lang}{os.path.splitext(args.filepath)[1]}")
except:
    sub.save_backup()
    traceback.print_exc()

translator.quit()
